/*
 * This program triggers an #OF exception and verifies that the instruction
 * pointer in the trap frame points to the instruction after the "INTO".
 *
 * In other words it makes sure that the #OF exception is a "trap-style"
 * exception.
 *
 * More debug information will be available on the console by setting
 * 'sysctl machdep.uprintf_signal' to '1'.
 */

#ifndef __i386__
#error "INTO only available in x86 32-bit mode"
#endif

#include <sys/types.h>
#include <sys/sysctl.h>

#include <x86/psl.h>
#include <machine/cpufunc.h>

#include <stdio.h>
#include <signal.h>
#include <assert.h>

void my_into(void);
void before_into(void) __asm(__STRING(before_into));
void after_into(void) __asm(__STRING(after_into));

static int got_fpe;
static int fpe_code;
static void *fpe_addr;

static void
fpe_handler(int num, siginfo_t *info, void *uap)
{
	got_fpe = 1;
	fpe_code = info->si_code;
	fpe_addr = info->si_addr;
}

int
main(void)
{
	u_long flags;
	int error;
	struct sigaction sa;

	sa.sa_sigaction = fpe_handler;
	sa.sa_flags = SA_SIGINFO;
	sigemptyset(&sa.sa_mask);
	error = sigaction(SIGFPE, &sa, NULL);
	assert(error == 0);

	my_into();

	printf("before_into: %p\n", before_into);
	printf("after_into: %p\n", after_into);
	printf("got_fpe: %d\n", got_fpe);
	printf("fpe_code: %d\n", fpe_code);
	printf("fpe_addr: %p\n", fpe_addr);
	printf("\n");

	if (got_fpe && fpe_code == FPE_INTOVF && fpe_addr == after_into)
		printf("success\n");
	else
		printf("failed\n");
}
